package com.example.eventbus;

import android.os.Handler;
import android.os.Looper;
import android.util.Log;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Create by cwb on 2019/2/18
 */
public class EventBus {

    private static volatile EventBus builder;

    private Map<Object, List<SubscribeMethod>> cacheMap;

    private Handler mHandler;

    private EventBus(){
        cacheMap = new HashMap<>();
        mHandler = new Handler();
    }

    public static EventBus getDefault() {
        if (builder == null) {
            synchronized (EventBus.class) {
                if (builder == null) {
                    builder = new EventBus();
                }
            }
        }
        return builder;
    }

    /**
     *  注册订阅者
     * @param subscriber 订阅者对象
     */
    public void register(Object subscriber) {
        List<SubscribeMethod> list = cacheMap.get(subscriber);
        if (list == null){
            list = findSubscribeMethod(subscriber);
            cacheMap.put(subscriber, list);
        }
    }

    private List<SubscribeMethod> findSubscribeMethod(Object object){
        List<SubscribeMethod> list = new ArrayList<>();
        Class<?> clazz = object.getClass();
        while (clazz != null){
            //过滤父类
            String name = clazz.getName();
            if (name.startsWith("java") || name.startsWith("android.")){
                break;
            }

            Method[] methods = clazz.getMethods();
            for (Method m: methods){
                Subscribe subscribe = m.getAnnotation(Subscribe.class);
                if (subscribe != null){
                    Class<?>[] types = m.getParameterTypes();
                    if (types.length != 1){
                        Log.e("Error:", "EventBus can only accept one param!");
                    }
                    ThreadMode threadMode = subscribe.threadMode();
                    SubscribeMethod subscribeMethod = new SubscribeMethod(m, threadMode, types[0]);
                    list.add(subscribeMethod);
                }
            }
            clazz = clazz.getSuperclass();
        }
        return list;
    }

    public void unregister(Object subscriber) {
        cacheMap.remove(subscriber);
    }

    /**
     *  发送数据
     * @param bean 实体类
     */
    public void post(final Object bean){
        Set<Object> set = cacheMap.keySet();
        Iterator<Object> iterator = set.iterator();
        while (iterator.hasNext()){
            final Object subscriber = iterator.next();
            List<SubscribeMethod> methods = cacheMap.get(subscriber);
            if (methods == null){
                continue;
            }
            for (final SubscribeMethod sMethod: methods){
                if (sMethod.getType().isAssignableFrom(bean.getClass())){
                    switch (sMethod.getThreadMode()){
                        case MAIN:
                            // 发送到主线程
                            if (Looper.myLooper() == Looper.getMainLooper()){
                                invoke(sMethod, subscriber, bean);
                            }else {
                                mHandler.post(new Runnable(){

                                    @Override
                                    public void run() {
                                        invoke(sMethod, subscriber, bean);
                                    }
                                });
                            }
                            break;
                        case BACKGROUND:
                            // 发送到子线程 ExecutorService

                            break;
                    }
                }
            }
        }
    }

    /**
     *
     * @param subscribeMethod  订阅方法
     * @param obj              订阅者对象
     * @param type             订阅方法接收的参数
     */
    private void invoke(SubscribeMethod subscribeMethod, Object obj, Object type){
        Method method = subscribeMethod.getMethod();
        try {
            method.invoke(obj, type);
        } catch (IllegalAccessException e) {
            e.printStackTrace();
        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }
    }
}
